#include <cstdio>
#include <cstring>
#include <screen.hpp>

/* Screen */

void Screen::resize(std::size_t x, std::size_t y) {
    w = x; h = y;
    buf = (Symbol *)realloc(buf, x * y * sizeof(Symbol));
}

Screen::Symbol &Screen::at(std::size_t n) { return buf[n]; }
Screen::Symbol &Screen::at(int x, int y) {
    auto idx = y * w + x;
    if(idx < 0 || idx >= w * h) return _dummy;
    return buf[idx];
}

Screen::Symbol &Screen::at(float x, float y) {
    auto [a, b] = xy_to_screen(x, y);
    if(a < 0 || a >= w || b < 0 || b >= h) return _dummy;
    return at(a, b);
}

std::pair<float, float> Screen::screen_to_xy(int x, int y) {
    return {
	x * sx + dx,
	y * sy + dy
    };
}

std::pair<int, int> Screen::xy_to_screen(float x, float y) {
    return {
	(x - dx)/sx,
	(y - dy)/sy
    };
}

void Screen::move_to(float x, float y) {
    dx = x - w*sx/2;
    dy = y - h*sy/2;
}

void Screen::clear() {
    static const Symbol s {' ', {0,0,0}};
    for(int i = 0; i < w * h; ++i)
	memcpy(buf + i, &s, sizeof(Symbol));
}

void Screen::draw() {
    // "shade" each pixel on screen with #this->shader()
    for(int i = 0; i < h; ++i) {
	for(int j = 0; j < w; ++j) {
	    auto [x, y] = screen_to_xy(j, i);
	    at(j, i) = shader(j, i, x, y);
	}
    }

    // draw lines for each velocity vector
    if(draw_vecs) {
	for(const auto &[start, vel] : vecs) {
	    auto end = *start + (*vel) / 3;
	    if(start->x == end.x) {
		// TODO: draw horizontal/vertical lines
		continue;
	    }

	    auto a = (end.y - start->y) / (end.x - start->x);
	    auto b = start->y - a * start->x;

	    for(int i = 0; i < w; ++i) {
		const auto [x, _] = screen_to_xy(i, 0);
		if(x < std::min(start->x, end.x)
		   || x > std::max(start->x, end.x)) continue;

		auto y = a*x + b;
		at(x, y) = Symbol{
		    .sym = '+',
		    .color = { 0.75, 0, 0 },
		};
	    }
	}
    }

    // write out a '#' wherever we have a point registered
    for(const auto &p : points) {
	at(p->x, p->y) = {.sym = '#', .color = 1};
    }

    // print out each symbol in our buffer to the terminal
    gotoxy(0, 0);
    for(int i = 0; i < h; ++i) {
	for(int j = 0; j < w; ++j) {
	    const auto &sym = at(j, i);
	    const auto color = sym.color.clamp({0,0,0}, {1,1,1}) * 255;
	    printf("\033[48;2;%d;%d;%dm%c",
		   (int)color.x, (int)color.y, (int)color.z, sym.sym);
	}

	// reset color, go to the next line
	printf("\033[38;2;%d;%d;%dm\n", 0,0,0);
    }

}

/* TanhColorizer */

vec<3> TanhColorizer::operator()(float y) {
    y = (std::tanh((y - translate)/scale * 2.25) + 1)/2.f;
    float hue = (1.f - y) * 240.f / 360.f;
    return hsl_to_rgb(hue, 1, 0.5);
}

vec<3> TanhColorizer::hsl_to_rgb(float h, float s, float l) {
    float q = l < 0.5 ? l * (1 + s) : l + s - l * s;
    float p = 2 * l - q;
    float r = hue_to_rgb(p, q, h + 0.33);
    float g = hue_to_rgb(p, q, h);
    float b = hue_to_rgb(p, q, h - 0.33);
    return { r, g, b };
}

float TanhColorizer::hue_to_rgb(float p, float q, float t) {
    if (t < 0) {
	t += 1;
    } else if (t > 1) {
	t -= 1;
    }

    if (t >= 0.66) {
	return p;
    } else if (t >= 0.5) {
	return p + (q - p) * (0.66 - t) * 6;
    } else if (t >= 0.33) {
	return q;
    } else {
	return p + (q - p) * 6 * t;
    }
}
